# -*- coding: utf-8 -*-
#
#  Copyright 2019 Ramil Nugmanov <nougmanoff@protonmail.com>
#  This file is part of CGRtools.
#
#  CGRtools is free software; you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with this program; if not, see <https://www.gnu.org/licenses/>.
#


class GroupI:
    __slots__ = ()


class GroupII:
    __slots__ = ()


class GroupIII:
    __slots__ = ()


class GroupIV:
    __slots__ = ()


class GroupV:
    __slots__ = ()


class GroupVI:
    __slots__ = ()


class GroupVII:
    __slots__ = ()


class GroupVIII:
    __slots__ = ()


class GroupIX:
    __slots__ = ()


class GroupX:
    __slots__ = ()


class GroupXI:
    __slots__ = ()


class GroupXII:
    __slots__ = ()


class GroupXIII:
    __slots__ = ()


class GroupXIV:
    __slots__ = ()


class GroupXV:
    __slots__ = ()


class GroupXVI:
    __slots__ = ()


class GroupXVII:
    __slots__ = ()


class GroupXVIII:
    __slots__ = ()
